@@ -103,12 +103,10 @@ completed_messagedialog_response_cb (GtkDialog *dialog,
 
 	case _OPEN_IN_BROWSER_RESPONSE:
 		{
-			GdkScreen    *screen;
 			OAuthAccount *account;
 			char         *url = NULL;
 			GError       *error = NULL;
 
-			screen = gtk_widget_get_screen (GTK_WIDGET (dialog));
 			gtk_widget_destroy (GTK_WIDGET (dialog));
 
 			account = web_service_get_current_account (WEB_SERVICE (data->service));
@@ -132,7 +130,7 @@ completed_messagedialog_response_cb (GtkDialog *dialog,
 			else if (data->photoset->id != NULL)
 				url = g_strconcat (data->server->url, "/photos/", account->id, "/sets/", data->photoset->id, NULL);
 
-			if ((url != NULL) && ! gtk_show_uri (screen, url, 0, &error)) {
+			if ((url != NULL) && ! gtk_show_uri_on_window (GTK_WINDOW (data->browser), url, GDK_CURRENT_TIME, &error)) {
 				if (data->service != NULL)
 					gth_task_dialog (GTH_TASK (data->service), TRUE, NULL);
 				_gtk_error_dialog_from_gerror_run (GTK_WINDOW (data->browser), _("Could not connect to the server"), error);
