@@ -27,6 +27,7 @@
 #include <gthumb.h>
 #include <extensions/image_viewer/image-viewer.h>
 #include <extensions/list_tools/list-tools.h>
+#include <extensions/list_tools/shortcuts.h>
 #include "actions.h"
 #include "callbacks.h"
 
@@ -42,15 +43,21 @@ static const GActionEntry actions[] = {
 };
 
 
+static const GthShortcut shortcuts[] = {
+	{ "rotate-right", N_("Rotate right"), GTH_SHORTCUT_CONTEXT_BROWSER_VIEWER, GTH_SHORTCUT_CATEGORY_LIST_TOOLS, "bracketright" },
+	{ "rotate-left", N_("Rotate left"), GTH_SHORTCUT_CONTEXT_BROWSER_VIEWER, GTH_SHORTCUT_CATEGORY_LIST_TOOLS, "bracketleft" },
+};
+
+
 static const GthMenuEntry tools1_action_entries[] = {
-	{ N_("Rotate Left"), "win.rotate-left", "bracketleft", "object-rotate-left-symbolic" },
-	{ N_("Rotate Right"), "win.rotate-right", "bracketright", "object-rotate-right-symbolic" },
+	{ N_("Rotate Left"), "win.rotate-left", NULL, "object-rotate-left-symbolic" },
+	{ N_("Rotate Right"), "win.rotate-right", NULL, "object-rotate-right-symbolic" },
 };
 
 
 static const GthMenuEntry tools2_action_entries[] = {
-	{ N_("Rotate Physically"), "win.apply-orientation", NULL },
-	{ N_("Reset the EXIF Orientation"), "win.reset-orientation", NULL }
+	{ N_("Rotate Physically"), "win.apply-orientation" },
+	{ N_("Reset the EXIF Orientation"), "win.reset-orientation" }
 };
 
 
@@ -84,6 +91,11 @@ ir__gth_browser_construct_cb (GthBrowser *browser)
 					 actions,
 					 G_N_ELEMENTS (actions),
 					 browser);
+
+	gth_window_add_shortcuts (GTH_WINDOW (browser),
+				  shortcuts,
+				  G_N_ELEMENTS (shortcuts));
+
 	gth_menu_manager_append_entries (gth_browser_get_menu_manager (browser, GTH_BROWSER_MENU_MANAGER_TOOLS),
 					 tools1_action_entries,
 					 G_N_ELEMENTS (tools1_action_entries));
@@ -94,12 +106,11 @@ ir__gth_browser_construct_cb (GthBrowser *browser)
 
 
 void
-ir__gth_browser_update_sensitivity_cb (GthBrowser *browser)
+ir__gth_browser_selection_changed_cb (GthBrowser *browser,
+				      int         n_selected)
 {
-	int      n_selected;
 	gboolean sensitive;
 
-	n_selected = gth_file_selection_get_n_selected (GTH_FILE_SELECTION (gth_browser_get_file_list_view (browser)));
 	sensitive = n_selected > 0;
 	gth_window_enable_action (GTH_WINDOW (browser), "rotate-right", sensitive);
 	gth_window_enable_action (GTH_WINDOW (browser), "rotate-left", sensitive);
@@ -108,31 +119,6 @@ ir__gth_browser_update_sensitivity_cb (GthBrowser *browser)
 }
 
 
-gpointer
-ir__gth_browser_file_list_key_press_cb (GthBrowser  *browser,
-					GdkEventKey *event)
-{
-	gpointer result = NULL;
-
-	switch (event->keyval) {
-	case GDK_KEY_bracketright:
-		gth_browser_activate_rotate_right (NULL, NULL, browser);
-		result = GINT_TO_POINTER (1);
-		break;
-
-	case GDK_KEY_bracketleft:
-		gth_browser_activate_rotate_left (NULL, NULL, browser);
-		result = GINT_TO_POINTER (1);
-		break;
-
-	default:
-		break;
-	}
-
-	return result;
-}
-
-
 static void
 viewer_image_changed_cb  (GtkWidget  *widget,
 		          GthBrowser *browser)
