@@ -40,8 +40,8 @@
 #define STEP_ANIMATION_BUTTON 5
 #define TRANSPARENCY_STYLE_BUTTON 6
 #undef ALWAYS_LOAD_ORIGINAL_SIZE
-#define N_FORWARD_PRELOADERS 2
-#define N_BACKWARD_PRELOADERS 2
+#define N_FORWARD_PRELOADERS 1
+#define N_BACKWARD_PRELOADERS 1
 
 
 static void gth_viewer_page_interface_init (GthViewerPageInterface *iface);
@@ -114,6 +114,8 @@ struct _GthImageViewerPagePrivate {
 	GthFileData       *next_file_data[N_FORWARD_PRELOADERS];
 	GthFileData       *prev_file_data[N_BACKWARD_PRELOADERS];
 	gulong             drag_data_get_event;
+	GCancellable      *current_cancellable;
+	GtkGesture        *gesture_zoom;
 };
 
 
@@ -175,7 +177,7 @@ _gth_image_preloader_get_requested_size_for_next_images (GthImageViewerPage *sel
 		requested_size = -1;
 		break;
 	default:
-		requested_size = (int) floor ((double) get_viewer_size (self) * 0.5);
+		requested_size = (int) get_viewer_size (self);
 		break;
 	}
 
@@ -245,8 +247,6 @@ _gth_image_viewer_page_load_with_preloader_step2 (GthImageViewerPage  *self,
 						  GAsyncReadyCallback  callback,
 						  gpointer	       user_data)
 {
-	g_object_ref (self);
-
 	gth_image_preloader_load (self->priv->preloader,
 				  file_data,
 				  requested_size,
@@ -255,9 +255,7 @@ _gth_image_viewer_page_load_with_preloader_step2 (GthImageViewerPage  *self,
 				  user_data,
 				  N_FORWARD_PRELOADERS + N_BACKWARD_PRELOADERS,
 				  self->priv->next_file_data[0],
-				  self->priv->next_file_data[1],
-				  self->priv->prev_file_data[0],
-				  self->priv->prev_file_data[1]);
+				  self->priv->prev_file_data[0]);
 }
 
 
@@ -328,35 +326,53 @@ _gth_image_viewer_page_load_with_preloader (GthImageViewerPage  *self,
 }
 
 
-static gboolean
-_gth_image_viewer_page_load_with_preloader_finish (GthImageViewerPage  *self)
-{
-	gboolean active = self->priv->active;
-	g_object_unref (self);
+typedef struct {
+	GthImageViewerPage *self;
+	GCancellable *cancellable;
+} LoadData;
 
-	return active;
+
+static LoadData *
+load_data_new (GthImageViewerPage *self)
+{
+	LoadData *load_data;
+
+	load_data = g_new0 (LoadData, 1);
+	load_data->self = g_object_ref (self);
+	load_data->cancellable = g_cancellable_new ();
+
+	return load_data;
+}
+
+
+static void
+load_data_free (LoadData *load_data)
+{
+	g_object_unref (load_data->self);
+	g_object_unref (load_data->cancellable);
+	g_free (load_data);
 }
 
 
 static void
 different_quality_ready_cb (GObject		*source_object,
 			    GAsyncResult	*result,
-			    gpointer	 	 user_data)
+			    gpointer		 user_data)
 {
-	GthImageViewerPage *self = user_data;
-	GthFileData	   *requested;
-	GthImage	   *image;
+	LoadData           *load_data = user_data;
+	GthImageViewerPage *self = load_data->self;
+	GthFileData	   *requested = NULL;
+	GthImage	   *image = NULL;
 	int		    requested_size;
 	int		    original_width;
 	int		    original_height;
 	GError		   *error = NULL;
-	cairo_surface_t    *s1 = NULL;
-	cairo_surface_t    *s2;
-	int                 w1, h1, w2, h2;
+	cairo_surface_t    *old_image;
+	cairo_surface_t    *new_image = NULL;
 	gboolean            got_better_quality;
 
-	if (! _gth_image_viewer_page_load_with_preloader_finish (self))
-		return;
+	if (self->priv->current_cancellable == load_data->cancellable)
+		self->priv->current_cancellable = NULL;
 
 	if (! gth_image_preloader_load_finish (GTH_IMAGE_PRELOADER (source_object),
 					       result,
@@ -368,31 +384,35 @@ different_quality_ready_cb (GObject		*source_object,
 					       &error))
 	{
 		g_clear_error (&error);
+		load_data_free (load_data);
 		return;
 	}
 
+	if (! self->priv->active)
+		goto clear_data;
+
 	if (! (self->priv->image_changed && requested == NULL) && ! _g_file_equal (requested->file, self->priv->file_data->file))
 		goto clear_data;
 
 	if (image == NULL)
 		goto clear_data;
 
-	/* check whether the image is of different quality */
+	/* Check whether the image is of different quality. */
 
-	s1 = gth_image_get_cairo_surface (image);
-	if (s1 == NULL)
+	new_image = gth_image_get_cairo_surface (image);
+	if (new_image == NULL)
 		goto clear_data;
 
-	s2 = gth_image_viewer_get_current_image (GTH_IMAGE_VIEWER (self->priv->viewer));
-	if (s2 == NULL) {
+	old_image = gth_image_viewer_get_current_image (GTH_IMAGE_VIEWER (self->priv->viewer));
+	if (old_image == NULL) {
 		got_better_quality = TRUE;
 	}
 	else {
-		w1 = cairo_image_surface_get_width (s1);
-		h1 = cairo_image_surface_get_height (s1);
-		w2 = cairo_image_surface_get_width (s2);
-		h2 = cairo_image_surface_get_height (s2);
-		got_better_quality = ((w1 > w2) || (h1 > h2));
+		int old_w = cairo_image_surface_get_width (old_image);
+		int old_h = cairo_image_surface_get_height (old_image);
+		int new_w = cairo_image_surface_get_width (new_image);
+		int new_h = cairo_image_surface_get_height (new_image);
+		got_better_quality = ((new_w > old_w) || (new_h > old_h));
 	}
 
 	if (got_better_quality) {
@@ -407,11 +427,11 @@ different_quality_ready_cb (GObject		*source_object,
 
 clear_data:
 
-	if (s1 != NULL)
-		cairo_surface_destroy (s1);
+	cairo_surface_destroy (new_image);
 	_g_object_unref (requested);
 	_g_object_unref (image);
 	g_clear_error (&error);
+	load_data_free (load_data);
 
 	return;
 }
@@ -440,13 +460,17 @@ _g_mime_type_can_load_different_quality (const char *mime_type)
 typedef struct {
 	GthImageViewerPage *self;
 	GthFileData *file_data;
+	guint        update_id;
 } UpdateQualityData;
 
 
 static void
 update_quality_data_free (UpdateQualityData *data)
 {
+	if (data->update_id == data->self->priv->update_quality_id)
+		data->self->priv->update_quality_id = 0;
 	_g_object_unref (data->file_data);
+	_g_object_unref (data->self);
 	g_free (data);
 }
 
@@ -458,42 +482,35 @@ update_quality_cb (gpointer user_data)
 	GthImageViewerPage *self = data->self;
 	gboolean            file_changed;
 
-	if (! _gth_image_viewer_page_load_with_preloader_finish (self)) {
-		update_quality_data_free (data);
-		return FALSE;
-	}
-
-	if (self->priv->update_quality_id != 0) {
-		g_source_remove (self->priv->update_quality_id);
-		self->priv->update_quality_id = 0;
-	}
+	if (! self->priv->active)
+		return G_SOURCE_REMOVE;
 
 	file_changed = ! _g_file_equal (data->file_data->file, self->priv->file_data->file);
-	update_quality_data_free (data);
-
 	if (file_changed)
-		return FALSE;
+		return G_SOURCE_REMOVE;
 
 	if (! self->priv->active)
-		return FALSE;
+		return G_SOURCE_REMOVE;
 
 	if (self->priv->viewer == NULL)
-		return FALSE;
+		return G_SOURCE_REMOVE;
 
-	if (self->priv->loading_image)
-		return FALSE;
+	if (self->priv->loading_image || (self->priv->current_cancellable != NULL))
+		return G_SOURCE_REMOVE;
 
 	if (! self->priv->image_changed && ! _g_mime_type_can_load_different_quality (gth_file_data_get_mime_type (self->priv->file_data)))
-		return FALSE;
+		return G_SOURCE_REMOVE;
 
+	LoadData *load_data = load_data_new (self);
+	self->priv->current_cancellable = load_data->cancellable;
 	_gth_image_viewer_page_load_with_preloader (self,
 						    self->priv->image_changed ? GTH_MODIFIED_IMAGE : self->priv->file_data,
 						    _gth_image_preloader_get_requested_size_for_current_image (self),
-						    NULL,
+						    load_data->cancellable,
 						    different_quality_ready_cb,
-						    self);
+						    load_data);
 
-	return FALSE;
+	return G_SOURCE_REMOVE;
 }
 
 
@@ -510,21 +527,24 @@ update_image_quality_if_required (GthImageViewerPage *self)
 	if ((image != NULL) && (gth_image_get_is_zoomable (image) || gth_image_get_is_animation (image)))
 		return;
 
+	if (self->priv->file_data == NULL)
+		return;
+
 	if (self->priv->update_quality_id != 0) {
 		g_source_remove (self->priv->update_quality_id);
 		self->priv->update_quality_id = 0;
 	}
 
 	UpdateQualityData *data;
-
 	data = g_new0 (UpdateQualityData, 1);
-	data->self = self;
+	data->self = _g_object_ref (self);
 	data->file_data = _g_object_ref (self->priv->file_data);
-
-	_g_object_ref (self);
-	self->priv->update_quality_id = g_timeout_add (UPDATE_QUALITY_DELAY,
-						       update_quality_cb,
-						       data);
+	data->update_id = g_timeout_add_full (G_PRIORITY_DEFAULT,
+					      UPDATE_QUALITY_DELAY,
+					      update_quality_cb,
+					      data,
+					      (GDestroyNotify) update_quality_data_free);
+	self->priv->update_quality_id = data->update_id;
 #endif
 }
 
@@ -588,8 +608,12 @@ update_overview_visibility (GthImageViewerPage *self)
 }
 
 
-#define MIN_ZOOM_LEVEL 0.3
-#define MAX_ZOOM_LEVEL 3.0
+#define MIN_ZOOM_LEVEL 0.05
+#define MAX_ZOOM_LEVEL 10.0
+#define RANGE_TO_ZOOM(x) (exp ((x) / 15.0 - M_E))
+#define MIN_RANGE_TO_ZOOM RANGE_TO_ZOOM(0.0)
+#define MAX_RANGE_TO_ZOOM RANGE_TO_ZOOM(100.0)
+#define ZOOM_TO_RANGE(x) ((log ((x)) + M_E) * 15.0)
 
 
 static void
@@ -600,7 +624,9 @@ zoom_scale_value_changed_cb (GtkScale *scale,
 	double              x, zoom;
 
 	x = gtk_range_get_value (GTK_RANGE (scale));
-	zoom = MIN_ZOOM_LEVEL + (x / 100.0 * (MAX_ZOOM_LEVEL - MIN_ZOOM_LEVEL));
+	x = (RANGE_TO_ZOOM (x) - MIN_RANGE_TO_ZOOM) / (MAX_RANGE_TO_ZOOM - MIN_RANGE_TO_ZOOM);
+	zoom = MIN_ZOOM_LEVEL + (x * (MAX_ZOOM_LEVEL - MIN_ZOOM_LEVEL));
+	zoom = CLAMP (zoom, MIN_ZOOM_LEVEL, MAX_ZOOM_LEVEL);
 	gth_image_viewer_set_zoom (GTH_IMAGE_VIEWER (self->priv->viewer), zoom);
 }
 
@@ -687,8 +713,11 @@ update_zoom_info (GthImageViewerPage *self)
 	GtkWidget *scale = _gtk_builder_get_widget (self->priv->builder, "zoom_level_scale");
 
 	_g_signal_handlers_block_by_data (scale, self);
-	x = (zoom - MIN_ZOOM_LEVEL) / (MAX_ZOOM_LEVEL - MIN_ZOOM_LEVEL) * 100.0;
-	gtk_range_set_value (GTK_RANGE (scale), CLAMP (x, 0, 100));
+	x = (zoom - MIN_ZOOM_LEVEL) / (MAX_ZOOM_LEVEL - MIN_ZOOM_LEVEL);
+	x = x * (MAX_RANGE_TO_ZOOM - MIN_RANGE_TO_ZOOM) + MIN_RANGE_TO_ZOOM;
+	x = ZOOM_TO_RANGE (x);
+	x = CLAMP (x, 0, 100);
+	gtk_range_set_value (GTK_RANGE (scale), x);
 	_g_signal_handlers_unblock_by_data (scale, self);
 }
 
@@ -733,6 +762,22 @@ viewer_popup_menu_cb (GtkWidget          *widget,
 }
 
 
+static void
+viewer_gesture_scale_changed_cb (GtkGestureZoom *controller,
+				 double          scale,
+				 gpointer        user_data)
+{
+
+	GthImageViewerPage *self = user_data;
+	double zoom = gth_image_viewer_get_zoom (GTH_IMAGE_VIEWER (self->priv->viewer));
+
+	scale = (scale > 1.0) ? 1.0 + ((scale - 1.0) * 0.125) : 1.0 - ((1.0 - scale) * 0.125);
+	zoom = zoom * scale;
+	zoom = CLAMP (zoom, MIN_ZOOM_LEVEL, MAX_ZOOM_LEVEL);
+	gth_image_viewer_set_zoom (GTH_IMAGE_VIEWER (self->priv->viewer), zoom);
+}
+
+
 static gboolean
 viewer_scroll_event_cb (GtkWidget 	   *widget,
 		        GdkEventScroll     *event,
@@ -1228,6 +1273,11 @@ gth_image_viewer_page_real_activate (GthViewerPage *base,
 				"map_event",
 				G_CALLBACK (viewer_image_map_event_cb),
 				self);
+	self->priv->gesture_zoom = gtk_gesture_zoom_new (self->priv->viewer);
+	g_signal_connect (self->priv->gesture_zoom,
+			  "scale-changed",
+			  G_CALLBACK (viewer_gesture_scale_changed_cb),
+			  self);
 	g_signal_connect (G_OBJECT (self->priv->viewer),
 			  "realize",
 			  G_CALLBACK (viewer_realize_cb),
@@ -1285,10 +1335,18 @@ gth_image_viewer_page_real_deactivate (GthViewerPage *base)
 		}
 	}
 
+	if (self->priv->current_cancellable != NULL) {
+		g_cancellable_cancel (self->priv->current_cancellable);
+		self->priv->current_cancellable = NULL;
+	}
+
 	_g_object_unref (self->priv->builder);
-	self->priv->builder = NULL;
 	_g_object_unref (self->priv->preloader);
+	gth_image_history_clear (self->priv->history);
+	_g_object_unref (self->priv->gesture_zoom);
+	self->priv->builder = NULL;
 	self->priv->preloader = NULL;
+	self->priv->gesture_zoom = NULL;
 	self->priv->active = FALSE;
 
 	gth_browser_set_viewer_widget (self->priv->browser, NULL);
@@ -1333,17 +1391,18 @@ preloader_load_ready_cb (GObject	*source_object,
 			 GAsyncResult	*result,
 			 gpointer	 user_data)
 {
-	GthImageViewerPage *self = user_data;
-	GthFileData	   *requested;
-	GthImage	   *image;
+	LoadData           *load_data = user_data;
+	GthImageViewerPage *self = load_data->self;
+	GthFileData	   *requested = NULL;
+	GthImage	   *image = NULL;
 	int		    requested_size;
 	int		    original_width;
 	int		    original_height;
 	GError		   *error = NULL;
 
 	self->priv->loading_image = FALSE;
-	if (! _gth_image_viewer_page_load_with_preloader_finish (self))
-		return;
+	if (self->priv->current_cancellable == load_data->cancellable)
+		self->priv->current_cancellable = NULL;
 
 	if (! gth_image_preloader_load_finish (GTH_IMAGE_PRELOADER (source_object),
 					       result,
@@ -1357,9 +1416,13 @@ preloader_load_ready_cb (GObject	*source_object,
 		if (! g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
 			gth_image_viewer_page_file_loaded (self, FALSE);
 		g_clear_error (&error);
+		load_data_free (load_data);
 		return;
 	}
 
+	if (!self->priv->active)
+		goto clear_data;
+
 	if (! _g_file_equal (requested->file, self->priv->file_data->file))
 		goto clear_data;
 
@@ -1420,14 +1483,13 @@ clear_data:
 	_g_object_unref (requested);
 	_g_object_unref (image);
 	g_clear_error (&error);
-
-	return;
+	load_data_free (load_data);
 }
 
 
 static void
 _gth_image_viewer_page_load (GthImageViewerPage *self,
-		             GthFileData        *file_data)
+			     GthFileData        *file_data)
 {
 	GthFileStore *file_store;
 	GtkTreeIter   iter;
@@ -1468,6 +1530,12 @@ _gth_image_viewer_page_load (GthImageViewerPage *self,
 		gth_image_viewer_set_void (GTH_IMAGE_VIEWER (self->priv->viewer));
 	}
 
+	if (self->priv->current_cancellable != NULL)
+		g_cancellable_cancel (self->priv->current_cancellable);
+
+	LoadData *load_data = load_data_new (self);
+	self->priv->current_cancellable = load_data->cancellable;
+
 	_gth_image_viewer_page_load_with_preloader (self,
 						    self->priv->file_data,
 #ifdef ALWAYS_LOAD_ORIGINAL_SIZE
@@ -1475,9 +1543,9 @@ _gth_image_viewer_page_load (GthImageViewerPage *self,
 #else
 						    _gth_image_preloader_get_requested_size_for_next_images (self),
 #endif
-						    NULL,
+						    load_data->cancellable,
 						    preloader_load_ready_cb,
-						    self);
+						    load_data);
 }
 
 
@@ -1519,6 +1587,16 @@ gth_image_viewer_page_real_focus (GthViewerPage *base)
 }
 
 
+static gboolean
+gth_image_viewer_page_real_has_focus (GthViewerPage *base)
+{
+	GtkWidget *widget;
+
+	widget = GTH_IMAGE_VIEWER_PAGE (base)->priv->viewer;
+	return gtk_widget_has_focus (widget);
+}
+
+
 static void
 gth_image_viewer_page_real_fullscreen (GthViewerPage *base,
 				       gboolean       active)
@@ -1808,6 +1886,7 @@ gth_image_viewer_page_real_save_as (GthViewerPage *base,
 
 	uri = g_file_get_uri (self->priv->file_data->file);
 	gtk_file_chooser_set_uri (GTK_FILE_CHOOSER (file_sel), uri);
+	gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (file_sel), g_file_info_get_edit_name (self->priv->file_data->info));
 
 	data = g_new0 (SaveAsData, 1);
 	data->self = self;
@@ -1961,6 +2040,30 @@ gth_image_viewer_page_shortcut_context (GthViewerPage *base)
 }
 
 
+static gboolean
+gth_image_viewer_page_can_open_clipboard (GthViewerPage *base,
+					  GdkAtom       *atoms,
+					  gint           n_atoms)
+{
+	// TODO: check atoms
+	return TRUE;
+}
+
+
+static void
+gth_image_viewer_page_open_clipboard (GthViewerPage *base,
+				      GtkClipboard  *clipboard)
+{
+	GthImageViewerPage *self = GTH_IMAGE_VIEWER_PAGE (base);
+	GthFileData        *file_data = gth_browser_get_current_file (self->priv->browser);
+	if (file_data == NULL)
+		return;
+	_g_object_unref (self->priv->file_data);
+	self->priv->file_data = gth_file_data_dup (file_data);
+	gth_image_viewer_page_paste_image (GTH_IMAGE_VIEWER_PAGE (base), clipboard);
+}
+
+
 static void
 gth_image_viewer_page_finalize (GObject *obj)
 {
@@ -2013,6 +2116,7 @@ gth_viewer_page_interface_init (GthViewerPageInterface *iface)
 	iface->can_view = gth_image_viewer_page_real_can_view;
 	iface->view = gth_image_viewer_page_real_view;
 	iface->focus = gth_image_viewer_page_real_focus;
+	iface->has_focus = gth_image_viewer_page_real_has_focus;
 	iface->fullscreen = gth_image_viewer_page_real_fullscreen;
 	iface->show_pointer = gth_image_viewer_page_real_show_pointer;
 	iface->update_sensitivity = gth_image_viewer_page_real_update_sensitivity;
@@ -2024,6 +2128,8 @@ gth_viewer_page_interface_init (GthViewerPageInterface *iface)
 	iface->zoom_from_scroll = gth_image_viewer_page_real_zoom_from_scroll;
 	iface->show_properties = gth_image_viewer_page_real_show_properties;
 	iface->shortcut_context = gth_image_viewer_page_shortcut_context;
+	iface->can_open_clipboard = gth_image_viewer_page_can_open_clipboard;
+	iface->open_clipboard = gth_image_viewer_page_open_clipboard;
 }
 
 
@@ -2060,6 +2166,7 @@ gth_image_viewer_page_init (GthImageViewerPage *self)
 		self->priv->prev_file_data[i] = NULL;
 
 	self->priv->drag_data_get_event = 0;
+	self->priv->current_cancellable = NULL;
 
 	/* settings notifications */
 
@@ -2333,11 +2440,11 @@ clipboard_image_received_cb (GtkClipboard *clipboard,
 
 
 void
-gth_image_viewer_page_paste_image (GthImageViewerPage *self)
+gth_image_viewer_page_paste_image (GthImageViewerPage *self,
+				   GtkClipboard       *clipboard)
 {
-	GtkClipboard *clipboard;
-
-	clipboard = gtk_clipboard_get_for_display (gtk_widget_get_display (self->priv->viewer), GDK_SELECTION_CLIPBOARD);
+	if (clipboard == NULL)
+		clipboard = gtk_clipboard_get_for_display (gtk_widget_get_display (self->priv->viewer), GDK_SELECTION_CLIPBOARD);
 	gtk_clipboard_request_image (clipboard,
 				     clipboard_image_received_cb,
 				     g_object_ref (self));
@@ -2389,12 +2496,6 @@ original_image_ready_cb (GObject	*source_object,
 	GthImage          *image = NULL;
 	GError            *error = NULL;
 
-	if (! _gth_image_viewer_page_load_with_preloader_finish (data->viewer_page)) {
-		g_task_return_error (data->task, g_error_new_literal (G_IO_ERROR, G_IO_ERROR_CANCELLED, ""));
-		get_original_data_free (data);
-		return;
-	}
-
 	if (! gth_image_preloader_load_finish (GTH_IMAGE_PRELOADER (source_object),
 					       result,
 					       NULL,
@@ -2405,9 +2506,14 @@ original_image_ready_cb (GObject	*source_object,
 					       &error))
 	{
 		g_task_return_error (data->task, error);
+		get_original_data_free (data);
+		return;
 	}
-	else
+
+	if (data->viewer_page->priv->active)
 		g_task_return_pointer (data->task, image, (GDestroyNotify) g_object_unref);
+	else
+		g_task_return_error (data->task, g_error_new_literal (G_IO_ERROR, G_IO_ERROR_CANCELLED, ""));
 
 	get_original_data_free (data);
 }
