@@ -594,6 +594,20 @@ _g_time_val_strftime (GTimeVal   *time_,
 }
 
 
+guint64
+_g_time_val_to_time (GTimeVal *time_)
+{
+	return ((guint64) (time_->tv_sec * G_USEC_PER_SEC)) + time_->tv_usec;
+}
+
+
+int
+_g_time_cmp (guint64 a, guint64 b)
+{
+	return (a == b) ? 0 : (a < b) ? -1 : 1;
+}
+
+
 /* Bookmark file utils */
 
 
@@ -1895,6 +1909,7 @@ _g_mime_type_is_video (const char *mime_type)
 
 	return (g_content_type_is_a (mime_type, "video/*")
 		|| (strcmp (mime_type, "application/ogg") == 0)
+		|| (strcmp (mime_type, "application/x-matroska") == 0)
 		|| (strcmp (mime_type, "application/vnd.ms-asf") == 0)
 		|| (strcmp (mime_type, "application/vnd.rn-realmedia") == 0));
 }
@@ -1921,7 +1936,29 @@ _g_mime_type_has_transparency (const char *mime_type)
 		|| (strcmp (mime_type, "image/jxl") == 0)
 		|| (strcmp (mime_type, "image/avif") == 0)
 		|| (strcmp (mime_type, "image/heif") == 0)
-		|| (strcmp (mime_type, "image/heic") == 0);
+		|| (strcmp (mime_type, "image/heic") == 0)
+		|| (strcmp (mime_type, "image/tiff") == 0)
+		|| (strcmp (mime_type, "image/x-tga") == 0);
+}
+
+
+char *
+_g_mime_type_get_generic_type (const char *mime_type)
+{
+	g_return_val_if_fail (mime_type != NULL, NULL);
+
+	const char *sep = strchr (mime_type, '/');
+	if (sep == NULL)
+		return NULL;
+
+	int sep_pos = sep - mime_type;
+	char *generic_type = g_new0 (char, sep_pos + 3);
+	strncpy (generic_type, mime_type, sep_pos + 1);
+	generic_type[sep_pos + 1] = '*';
+	generic_type[sep_pos + 2] = 0;
+
+	return generic_type;
+
 }
 
 
